///////////////////////////////////////////////////////////////////////
// File:  CRC16_D16.v                             
// Date:  Fri May 23 11:10:11 2008                                                      
//                                                                     
// Copyright (C) 1999-2003 Easics NV.                 
// This source file may be used and distributed without restriction    
// provided that this copyright statement is not removed from the file 
// and that any derivative work contains the original copyright notice
// and the associated disclaimer.
//
// THIS SOURCE FILE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Purpose: Verilog module containing a synthesizable CRC function
//   * polynomial: (0 2 15 16)
//   * data width: 16
//                                                                     
// Info: tools@easics.be
//       http://www.easics.com                                  
///////////////////////////////////////////////////////////////////////
  // convention: the first serial data bit is D[15]
/* http://www.easics.be/webtools/crctool with Polynomial:  1 + x2 + x15 + x16
Changes from original file: added `include, removed function, introduced clock. - TG*/

`include "define.v"
module CRC16D16   (
 input  clk, init,
 input  [15:0] DIn,
 output [15:0] CRC  );

reg  [15:0] C;
wire [15:0] D;
assign D = DIn;
assign CRC = C;

always @ (posedge clk) 
 if (init) C <= 16'hFFFF;
 else begin
   C[0]  <= D[15] ^ D[13] ^ D[12] ^ D[11] ^ D[10] ^ D[9] ^ D[8] ^ 
           D[7] ^ D[6] ^ D[5] ^ D[4] ^ D[3] ^ D[2] ^ D[1] ^ D[0] ^ 
           C[0] ^ C[1] ^ C[2] ^ C[3] ^ C[4] ^ C[5] ^ C[6] ^ C[7] ^ 
           C[8] ^ C[9] ^ C[10] ^ C[11] ^ C[12] ^ C[13] ^ C[15];
   C[1]  <= D[14] ^ D[13] ^ D[12] ^ D[11] ^ D[10] ^ D[9] ^ D[8] ^ 
           D[7] ^ D[6] ^ D[5] ^ D[4] ^ D[3] ^ D[2] ^ D[1] ^ C[1] ^ 
           C[2] ^ C[3] ^ C[4] ^ C[5] ^ C[6] ^ C[7] ^ C[8] ^ C[9] ^ 
           C[10] ^ C[11] ^ C[12] ^ C[13] ^ C[14];
   C[2]  <= D[14] ^ D[1] ^ D[0] ^ C[0] ^ C[1] ^ C[14];
   C[3]  <= D[15] ^ D[2] ^ D[1] ^ C[1] ^ C[2] ^ C[15];
   C[4]  <= D[3] ^ D[2] ^ C[2] ^ C[3];
   C[5]  <= D[4] ^ D[3] ^ C[3] ^ C[4];
   C[6]  <= D[5] ^ D[4] ^ C[4] ^ C[5];
   C[7]  <= D[6] ^ D[5] ^ C[5] ^ C[6];
   C[8]  <= D[7] ^ D[6] ^ C[6] ^ C[7];
   C[9]  <= D[8] ^ D[7] ^ C[7] ^ C[8];
   C[10] <= D[9] ^ D[8] ^ C[8] ^ C[9];
   C[11] <= D[10] ^ D[9] ^ C[9] ^ C[10];
   C[12] <= D[11] ^ D[10] ^ C[10] ^ C[11];
   C[13] <= D[12] ^ D[11] ^ C[11] ^ C[12];
   C[14] <= D[13] ^ D[12] ^ C[12] ^ C[13];
   C[15] <= D[15] ^ D[14] ^ D[12] ^ D[11] ^ D[10] ^ D[9] ^ D[8] ^ 
           D[7] ^ D[6] ^ D[5] ^ D[4] ^ D[3] ^ D[2] ^ D[1] ^ D[0] ^ 
           C[0] ^ C[1] ^ C[2] ^ C[3] ^ C[4] ^ C[5] ^ C[6] ^ C[7] ^ 
           C[8] ^ C[9] ^ C[10] ^ C[11] ^ C[12] ^ C[14] ^ C[15];

  end // else  

endmodule

