library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;
library UNISIM;
use UNISIM.VComponents.all;
entity LVDS is
    Port ( l_clk : in std_logic;
           strobe : in std_logic;
           clk : in std_logic;
           hq : in std_logic_vector(19 downto 0);
           cc_rst : in std_logic;
           gc_block : in std_logic;
           gc_tstrip : in std_logic;
           gc_idmatch : in std_logic;
           gc_automon : in std_logic;
           ch_number : in std_logic_vector(4 downto 0);
           lfifo_ack : in std_logic;
					 trailer : out std_logic;
					 lfifo_req : out std_logic;
					 lfifo_we : out std_logic;
           lfifo_data : out std_logic_vector(32 downto 0);
-- monitoring counting signals
					 update : out std_logic;
					 BADID : out std_logic;
					 UERR : out std_logic;
					 CERR : out std_logic
			  );
end LVDS;
architecture my_arch of LVDS is
type state is (idle,head,dat1,dat0,even,trlr);
signal RCV : state;
signal e_odd : std_logic := '0';
signal e_struct : std_logic := '0';
signal bad_id : std_logic := '0';
signal IDREG : std_logic_vector(7 downto 0) := (others =>'0');
signal erh : std_logic_vector(1 downto 0) := (others =>'0');
signal data_cntr : std_logic_vector(2 downto 0) := (others =>'0');
signal idle_cntr : std_logic_vector(2 downto 0) := (others =>'0');
-- LVDS fifo signals
signal preg : std_logic_vector(15 downto 0) := (others =>'0');
signal block_wc : std_logic_vector(15 downto 0) := (others =>'0');
signal we_lfifo : std_logic_vector(1 downto 0) := (others =>'0');
signal lfifo_wa : std_logic_vector(9 downto 0) := (others =>'0');
signal lfifo_ra : std_logic_vector(8 downto 0) := (others =>'0');
signal lfifo_wc : std_logic_vector(8 downto 0) := (others =>'0');
signal lfifo_di : std_logic_vector(17 downto 0) := (others =>'0');
signal lfifo_do : std_logic_vector(35 downto 0) := (others =>'0');
signal wa_sync : std_logic_vector(2 downto 0) := (others =>'0');
signal uerr_sync : std_logic_vector(2 downto 0) := (others =>'0');
signal cerr_sync : std_logic_vector(2 downto 0) := (others =>'0');
signal erh_toggle : std_logic_vector(1 downto 0) := (others =>'0');
signal wr_odd : std_logic := '0';
signal up_count : std_logic := '0';
signal ec_lfifo_ra : std_logic := '0';
signal lfifo_we_i : std_logic := '0';
signal header : std_logic := '0';
signal trailer_i : std_logic := '0';
signal updc : std_logic := '0';
signal rst_lfifo_do : std_logic := '0';
signal event_end : std_logic := '0';
signal blk_length : std_logic_vector(6 downto 0) := (others =>'0');
signal trailer_sync : std_logic_vector(2 downto 0) := (others =>'0');
begin
lfifo_we <= lfifo_we_i;
trailer <= trailer_i;
-- state machine
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    RCV <= idle;
  elsif(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			if(RCV = idle and hq(17) = '1')then
				RCV <= head;
			end if;
			if(RCV = head)then
				RCV <= dat1;
			end if;
			if(RCV = dat1)then
				if((hq(19) = '0' and hq(17 downto 16) = "01") or idle_cntr(2) = '1')then
					RCV <= even;
				elsif(hq(17) = '1')then
					RCV <= dat0;
				end if;
			end if;
			if(RCV = dat0)then
				if((hq(19) = '0' and hq(17 downto 16) = "01") or idle_cntr(2) = '1')then
					RCV <= trlr;
				else
					RCV <= dat1;
				end if;
			end if;
			if(RCV = even)then
				RCV <= trlr;
			end if;
			if(RCV = trlr)then
				if(hq(17 downto 16) = "11")then
					RCV <= head;
				else
					RCV <= idle;
				end if;
			end if;
		end if;
  end if;
end process;
-- err status
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    e_struct <= '0';
    e_odd <= '0';
    erh <= "00";
  elsif(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			if(RCV = trlr)then
				e_struct <= '0';
				e_odd <= '0';
				erh <= "00";
			else
				if(RCV = idle and hq(17 downto 16) = "10")then
					e_struct <= '1';
				end if;
				if(RCV = head and hq(17 downto 16) /= "10")then
					e_struct <= '1';
				end if;
				if(RCV = dat1 and hq(16) = '1')then
					e_struct <= '1';
				end if;
				if(RCV = dat0 and hq(16) = hq(17))then
					e_struct <= '1';
				end if;
				if(RCV = even)then
					e_odd <= '1';
				end if;
				erh <= erh or hq(19 downto 18);
			end if;
		end if;
  end if;
end process;
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    erh_toggle <= "00";
  elsif(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			erh_toggle <= erh_toggle xor hq(19 downto 18);
		end if;
  end if;
end process;
-- LVDS fifo
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    we_lfifo(0) <= '0';
    wr_odd <= '0';
  elsif(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			if(RCV = head or RCV = even or RCV = dat0)then
				we_lfifo(0) <= '1';
				wr_odd <= '0';
			else
				we_lfifo(0) <= we_lfifo(0) and not wr_odd;
				wr_odd <= wr_odd xor we_lfifo(0);
			end if;
		end if;
  end if;
end process;
we_lfifo(1) <= we_lfifo(0);
process(l_clk)
begin
  if(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			if(RCV = idle or hq(17) = '1')then
				idle_cntr <= "000";
			else
				idle_cntr <= idle_cntr + 1;
			end if;
			if(RCV = head)then
			 IDREG <= preg(7 downto 0);
			end if;
			if(RCV = head)then
				block_wc <= x"0001";
			elsif(we_lfifo(0) = '1' and wr_odd = '0')then
				block_wc <= block_wc + gc_block;
			end if;
			if(RCV = head)then
				lfifo_di(15 downto 0) <= preg(15 downto 13) & ch_number & preg(7 downto 0);
			elsif(RCV = trlr and gc_tstrip = '0' )then
				lfifo_di(15 downto 0) <= preg(15 downto 8) & e_odd & e_struct & "00" & bad_id & '0' & erh;
			elsif((RCV = even or (HQ(16) = '1' and RCV = dat0)) and gc_block = '1')then
				lfifo_di(15 downto 0) <= block_wc;
			else
				lfifo_di(15 downto 0) <= preg;
			end if;
			if(RCV = trlr)then
				lfifo_di(16) <= '1';
			else
				lfifo_di(16) <= '0';
			end if;
			preg <= hq(15 downto 0);
			if(hq(15 downto 8) /= IDREG and gc_idmatch = '1')then
				bad_id <= '1';
			else
				bad_id <= '0';
			end if;
		end if;
  end if;
end process;
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    lfifo_wa <= (others => '0');
  elsif(l_clk'event and l_clk = '1')then
    if(strobe = '1')then
			if(we_lfifo(0) = '1')then
				lfifo_wa <= lfifo_wa + 1;
			end if;
		end if;
  end if;
end process;
lfifo_di(17) <= '0';
i_lfifo : RAMB16BWE_S36_S18
   port map (
      DOA => lfifo_do(31 downto 0),      -- Port A 32-bit Data Output
      DOPA => lfifo_do(35 downto 32),    -- Port A 4-bit Parity Output
      ADDRA => lfifo_ra,  -- Port A 9-bit Address Input
      ADDRB => lfifo_wa,  -- Port B 10-bit Address Input
      CLKA => clk,    -- Port A 1-bit Clock
      CLKB => l_clk,    -- Port B 1-bit Clock
      DIA => x"00000000",      -- Port A 32-bit Data Input
      DIB => lfifo_di(15 downto 0),      -- Port B 16-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => lfifo_di(17 downto 16),    -- Port B 2-bit parity Input
      ENA => '1',      -- Port A 1-bit RAM Enable Input
      ENB => strobe,      -- Port B 1-bit RAM Enable Input
      SSRA => rst_lfifo_do,    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => x"0",      -- Port A 4-bit Write Enable Input
      WEB => we_lfifo       -- Port B 2-bit Write Enable Input
   );
rst_lfifo_do <= not ec_lfifo_ra;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(cc_rst = '1')then
			uerr_sync <= "000";
			cerr_sync <= "000";
			UERR <= '0';
			CERR <= '0';
		else	
			uerr_sync <= uerr_sync(1 downto 0) & erh_toggle(1);
			cerr_sync <= cerr_sync(1 downto 0) & erh_toggle(0);
			UERR <= uerr_sync(2) xor uerr_sync(1);
			CERR <= cerr_sync(2) xor cerr_sync(1);
		end if;
	end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(cc_rst = '1')then
			wa_sync <= (others => '0');
			up_count <= '0';
			lfifo_ra <= (others => '0');
			lfifo_wc <= (others => '0');
		else
			wa_sync <= wa_sync(1 downto 0) & lfifo_wa(1);
			up_count <= wa_sync(2) xor wa_sync(1);
			if(ec_lfifo_ra = '1')then
				lfifo_ra <= lfifo_ra + 1;
			end if;	
			if(up_count /= ec_lfifo_ra)then
				if(up_count = '1')then
				  lfifo_wc <= lfifo_wc + 1;
				else
				  lfifo_wc <= lfifo_wc - 1;
				end if;
			end if;
		end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(cc_rst = '1')then
			lfifo_req <= '0';
			header <= '1';
			updc <= '0';
			update <= '0';
			BADID <= '0';
			trailer_i <= '0';
			ec_lfifo_ra <= '0';
			lfifo_we_i <= '0';
			trailer_sync <= "000";
		else
			if(ec_lfifo_ra = '1')then
				lfifo_req <= '0';
			elsif(or_reduce(lfifo_wc(8 downto 3)) = '1' or (event_end = '1' and lfifo_wc(2 downto 0) /= "000"))then
				lfifo_req <= '1';
			end if;
			header <= trailer_i or (header and not lfifo_we_i);
	-- if bit(15) of header = '1', update channel registers
--			if(header = '1' and lfifo_we_i = '1')then
			if(header = '1')then
				updc <= lfifo_do(15);
			end if;
			update <= not gc_automon and updc and trailer_i;
--			BADID <= lfifo_do(3) and lfifo_do(34) and lfifo_we_i;
			BADID <= lfifo_do(3) and lfifo_do(34);
--			trailer_i <= lfifo_do(34) and lfifo_we_i;
			trailer_i <= lfifo_do(34);
			if(lfifo_ack = '1')then
			  ec_lfifo_ra <= '1';
			elsif((lfifo_wc(8 downto 1) = x"00" and (lfifo_wc(0) = '0' or ec_lfifo_ra = '1')) or blk_length(6) = '1')then
			  ec_lfifo_ra <= '0';
			end if;
			lfifo_we_i <= ec_lfifo_ra;
			trailer_sync <= trailer_sync(1 downto 0) & lfifo_di(16);
			if(lfifo_do(34) = '1')then
			  event_end <= '0';
			elsif(trailer_sync(2 downto 1) = "10")then
			  event_end <= '1';
			end if;
		end if;
		if(ec_lfifo_ra = '0')then
			blk_length <= "0000001";
		else
			blk_length <= blk_length + 1;
		end if;
  end if;  
end process;
lfifo_data <= lfifo_do(34) & lfifo_do(31 downto 0);
end my_arch;
