library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
entity Hamming is
    Port ( clk : in std_logic;
           strobe : in std_logic;
           din : in std_logic_vector(23 downto 0);
					 cc_rxena : in std_logic;
           cc_ecc : in std_logic;
           dout : out std_logic_vector(19 downto 0)
           );
end Hamming;

architecture my_arch of Hamming is
signal d: std_logic_vector(24 downto 1) := (others =>'0');
signal q: std_logic_vector(18 downto 1) := (others =>'0');
signal c: std_logic_vector(5 downto 1) := (others =>'0');
signal pp: std_logic := '0';
signal strobe_i: std_logic_vector(2 downto 0) := (others =>'0');
signal doutp: std_logic_vector(19 downto 0) := (others =>'0');
signal dout_i: std_logic_vector(19 downto 0) := (others =>'0');
begin
d <= din(23) & din(13) & '0' & din(20 downto 14) & din(22) & din(12 downto 6) &  din(21) & din(4 downto 0);
process(clk)
begin
  if(clk'event and clk = '1')then
		strobe_i(2) <= strobe;
		strobe_i(1) <= strobe;
		strobe_i(0) <= strobe;
		if(strobe_i(0) = '1')then
			c(1) <= d(1) xor d(3) xor d(5) xor d(7) xor d(9) xor d(11) xor
							 d(13) xor d(15) xor d(17) xor d(19) xor d(21) xor d(23);
			c(2) <= d(2) xor d(3) xor d(6) xor d(7) xor d(10) xor d(11) xor
							 d(14) xor d(15) xor d(18) xor d(19) xor d(22) xor d(23);
			c(3) <= d(4) xor d(5) xor d(6) xor d(7) xor d(12) xor d(13) xor
							 d(14) xor d(15) xor d(20) xor d(21) xor d(22) xor d(23);
			c(4) <= d(8) xor d(9) xor d(10) xor d(11) xor d(12) xor d(13) xor
							 d(14) xor d(15);
			c(5) <= d(16) xor d(17) xor d(18) xor d(19) xor d(20) xor d(21) xor
							 d(22) xor d(23);
			pp   <= d(2) xor d(4) xor d(6) xor d(8) xor d(10) xor d(12) xor
							 d(14) xor d(16) xor d(18) xor d(20) xor d(22) xor d(24);
			q(1) <= d(3);
			q(2) <= d(5);
			q(3) <= d(6);
			q(4) <= d(7);
			q(5) <= d(9);
			q(6) <= d(10);
			q(7) <= d(11);
			q(8) <= d(12);
			q(9) <= d(13);
			q(10) <= d(14);
			q(11) <= d(15);
			q(12) <= d(17);
			q(13) <= d(18);
			q(14) <= d(19);
			q(15) <= d(20);
			q(16) <= d(21);
			q(17) <= d(22);
			q(18) <= d(23);
		end if;
  end if;
end process;
process(clk,cc_rxena)
begin
  if(cc_rxena = '0')then
    doutp <= (others => '0');
    dout_i <= (others => '0');
  elsif(clk'event and clk = '1')then
		if(strobe_i(1) = '1')then
			if(pp = '0' and c = "00011" and cc_ecc = '1')then
				doutp(0) <= not q(1);
			else
				doutp(0) <= q(1);
			end if;
			if(pp = '0' and c = "00101" and cc_ecc = '1')then
				doutp(1) <= not q(2);
			else
				doutp(1) <= q(2);
			end if;
			if(pp = '1' and c = "00110" and cc_ecc = '1')then
				doutp(2) <= not q(3);
			else
				doutp(2) <= q(3);
			end if;
			if(pp = '0' and c = "00111" and cc_ecc = '1')then
				doutp(3) <= not q(4);
			else
				doutp(3) <= q(4);
			end if;
			if(pp = '0' and c = "01001" and cc_ecc = '1')then
				doutp(4) <= not q(5);
			else
				doutp(4) <= q(5);
			end if;
			if(pp = '1' and c = "01010" and cc_ecc = '1')then
				doutp(5) <= not q(6);
			else
				doutp(5) <= q(6);
			end if;
			if(pp = '0' and c = "01011" and cc_ecc = '1')then
				doutp(6) <= not q(7);
			else
				doutp(6) <= q(7);
			end if;
			if(pp = '1' and c = "01100" and cc_ecc = '1')then
				doutp(7) <= not q(8);
			else
				doutp(7) <= q(8);
			end if;
			if(pp = '0' and c = "01101" and cc_ecc = '1')then
				doutp(8) <= not q(9);
			else
				doutp(8) <= q(9);
			end if;
			if(pp = '1' and c = "01110" and cc_ecc = '1')then
				doutp(9) <= not q(10);
			else
				doutp(9) <= q(10);
			end if;
			if(pp = '0' and c = "01111" and cc_ecc = '1')then
				doutp(10) <= not q(11);
			else
				doutp(10) <= q(11);
			end if;
			if(pp = '0' and c = "10001" and cc_ecc = '1')then
				doutp(11) <= not q(12);
			else
				doutp(11) <= q(12);
			end if;
			if(pp = '1' and c = "10010" and cc_ecc = '1')then
				doutp(12) <= not q(13);
			else
				doutp(12) <= q(13);
			end if;
			if(pp = '0' and c = "10011" and cc_ecc = '1')then
				doutp(13) <= not q(14);
			else
				doutp(13) <= q(14);
			end if;
			if(pp = '1' and c = "10100" and cc_ecc = '1')then
				doutp(14) <= not q(15);
			else
				doutp(14) <= q(15);
			end if;
			if(pp = '0' and c = "10101" and cc_ecc = '1')then
				doutp(15) <= not q(16);
			else
				doutp(15) <= q(16);
			end if;
			if(pp = '1' and c = "10111" and cc_ecc = '1')then
				doutp(17) <= not q(18);
			else
				doutp(17) <= q(18);
			end if;
			doutp(18) <= c(1) xor pp;
			if(c /= "00000" and c(1) = pp )then
				doutp(19) <= '1';
			else
				doutp(19) <= '0';
			end if;
		end if;
		if(strobe_i(2) = '1')then
			dout_i(15 downto 0) <= doutp(15 downto 0);
			dout_i(19 downto 17) <= doutp(19 downto 17);
			dout_i(16) <= doutp(17) xor dout_i(17);
		end if;
  end if;
end process;
dout <= dout_i;
end my_arch;
