library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;
entity LVDS is
    Port ( l_clk : in std_logic;
           l_strobe : in std_logic;
           strobe : in std_logic;
           clk : in std_logic;
           hq : in std_logic_vector(19 downto 0);
           cc_rst : in std_logic;
--           gc_block : in std_logic;
--           gc_tstrip : in std_logic;
--           gc_idmatch : in std_logic;
--           gc_automon : in std_logic;
           ch_number : in std_logic_vector(4 downto 0);
           lfifo_ack : in std_logic;
--					 trailer : out std_logic;
					 lfifo_req : out std_logic;
					 lfifo_we : out std_logic;
           lfifo_data : out std_logic_vector(32 downto 0);
-- monitoring counting signals
--					 update : out std_logic;
--					 BADID : out std_logic;
					 UERR : out std_logic;
					 CERR : out std_logic
			  );
end LVDS;
architecture my_arch of LVDS is
component cntr
  port (
    reset: in std_logic;
    ec: in std_logic;
    clk: in std_logic;
    q: out std_logic_vector (2 downto 0)
  );
end component;
type state is (idle,head,dat1,dat0,even,trlr);
signal RCV : state;
signal strobe_i: std_logic_vector(3 downto 0) := (others =>'0');
signal e_odd : std_logic := '0';
signal e_trunc : std_logic := '0';
signal e_struct : std_logic := '0';
signal bad_id : std_logic := '0';
signal IDREG : std_logic_vector(7 downto 0) := (others =>'0');
signal erh : std_logic_vector(1 downto 0) := (others =>'0');
--signal data_cntr : std_logic_vector(2 downto 0) := (others =>'0');
signal idle_cntr : std_logic_vector(2 downto 0) := (others =>'0');
signal rst_ic : std_logic := '0';
-- LVDS fifo signals
signal crc : std_logic_vector(15 downto 0) := (others =>'0');
signal init_crc : std_logic := '0';
signal e_crc : std_logic := '0';
signal e_short : std_logic := '0';
signal preg : std_logic_vector(15 downto 0) := (others =>'0');
signal block_wc : std_logic_vector(9 downto 0) := (others =>'0');
signal we_lfifo : std_logic_vector(3 downto 0) := (others =>'0');
signal lfifo_wa : std_logic_vector(8 downto 0) := (others =>'0');
signal lfifo_ra : std_logic_vector(8 downto 0) := (others =>'0');
signal lfifo_wc : std_logic_vector(8 downto 0) := (others =>'0');
signal lfifo_di : std_logic_vector(35 downto 0) := (others =>'0');
signal lfifo_do : std_logic_vector(35 downto 0) := (others =>'0');
signal wa_sync : std_logic_vector(2 downto 0) := (others =>'0');
signal rqst_cntr : std_logic_vector(3 downto 0) := (others =>'0');
signal up_count : std_logic := '0';
signal ec_lfifo_ra : std_logic := '0';
signal lfifo_we_i : std_logic := '0';
signal lfifo_req_i : std_logic := '0';
--signal trailer_i : std_logic := '0';
signal rst_lfifo_do : std_logic := '0';
signal header : std_logic := '0';
signal blk_length : std_logic_vector(6 downto 0) := (others =>'0');
begin
lfifo_we <= lfifo_we_i;
lfifo_req <= lfifo_req_i;
--trailer <= trailer_i;
-- state machine
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    RCV <= idle;
  elsif(l_clk'event and l_clk = '1')then
    if(strobe_i(0) = '1')then
			if(RCV = idle and hq(17) = '1')then
				RCV <= head;
			end if;
			if(RCV = head)then
				RCV <= dat1;
			end if;
			if(RCV = dat1)then
				if((hq(19) = '0' and hq(17 downto 16) = "01") or idle_cntr(2) = '1')then
					RCV <= even;
				elsif(hq(17) = '1')then
					RCV <= dat0;
				end if;
			end if;
			if(RCV = dat0)then
				if((hq(19) = '0' and hq(17 downto 16) = "01") or idle_cntr(2) = '1')then
					RCV <= trlr;
				else
					RCV <= dat1;
				end if;
			end if;
			if(RCV = even)then
				RCV <= trlr;
			end if;
			if(RCV = trlr)then
				if(hq(17 downto 16) = "11")then
					RCV <= head;
				else
					RCV <= idle;
				end if;
			end if;
		end if;
  end if;
end process;
i_crc16D16: crc16D16 PORT MAP(
		clk => l_clk,
		init_crc => init_crc,
		we_crc => strobe_i(1),
		d => preg,
		crc => crc
	);
init_crc <= '1' when RCV = head else '0';
-- err status
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    e_struct <= '0';
    e_odd <= '0';
    erh <= "00";
  elsif(l_clk'event and l_clk = '1')then
    if(strobe_i(0) = '1')then
			if(RCV = trlr)then
				e_struct <= '0';
				e_odd <= '0';
				erh <= "00";
			else
				if(RCV = idle and hq(17 downto 16) = "10")then
					e_struct <= '1';
				end if;
				if(RCV = head and hq(17 downto 16) /= "10")then
					e_struct <= '1';
				end if;
				if(RCV = dat1 and hq(16) = '1')then
					e_struct <= '1';
				end if;
				if(RCV = dat0 and hq(16) = hq(17))then
					e_struct <= '1';
				end if;
				if(RCV = even)then
					e_odd <= '1';
				end if;
				if(RCV /= idle)then
					erh <= erh or hq(19 downto 18);
				end if;
			end if;
		end if;
  end if;
end process;
-- LVDS fifo
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    we_lfifo(0) <= '0';
    we_lfifo(2) <= '0';
  elsif(l_clk'event and l_clk = '1')then
    if(strobe_i(0) = '1')then
			if(RCV = head or (RCV = even or RCV = dat0))then
				we_lfifo(0) <= '1';
			else
				we_lfifo(0) <= '0';
			end if;
			if(block_wc(9) = '0' or RCV = trlr)then
				we_lfifo(2) <= we_lfifo(0);
			else
				we_lfifo(2) <= '0';
			end if;
		end if;
  end if;
end process;
we_lfifo(1) <= we_lfifo(0);
we_lfifo(3) <= we_lfifo(2);
rst_ic <= '1' when RCV = idle or hq(17) = '1' else '0';
i_idle_cntr: cntr port map(reset => rst_ic, clk => l_clk, ec => strobe_i(2), q => idle_cntr);
process(l_clk)
begin
  if(l_clk'event and l_clk = '1')then
		strobe_i(0) <= l_strobe;
		strobe_i(1) <= l_strobe;
		strobe_i(2) <= l_strobe;
		strobe_i(3) <= l_strobe;
    if(strobe_i(2) = '1')then
--			if(RCV = idle or hq(17) = '1')then
--				idle_cntr <= "000";
--			else
--				idle_cntr <= idle_cntr + 1;
--			end if;
			if(RCV = head)then
			 IDREG <= preg(7 downto 0);
			end if;
			if(RCV = head)then
				block_wc <= "0000000001";
			elsif(we_lfifo(0) = '1' and block_wc(9) = '0')then
				block_wc <= block_wc + 1;
			end if;
			if(we_lfifo(0) = '1')then
				e_trunc <= block_wc(9);
			end if;
			if(crc = preg)then
				e_crc <= '0';
			else
				e_crc <= '1';
			end if;
			if(RCV = head)then
				e_short <= '1';
			elsif(block_wc(2) = '1')then
				e_short <= '0';
			end if;
			if(RCV = head)then
				lfifo_di(15 downto 0) <= preg(15 downto 13) & ch_number & preg(7 downto 0);
			elsif(RCV = trlr)then
				lfifo_di(15 downto 0) <= preg(15 downto 8) & e_odd & e_struct & e_short & e_crc & bad_id & e_trunc & erh;
			elsif(RCV = even or (hq(16) = '1' and RCV = dat0))then
				lfifo_di(15 downto 0) <= "000000" & block_wc;
			else
				lfifo_di(15 downto 0) <= preg;
			end if;
		end if;
    if(strobe_i(3) = '1')then
			preg <= hq(15 downto 0);
			if(hq(15 downto 8) /= IDREG)then
				bad_id <= '1';
			else
				bad_id <= '0';
			end if;
			if(we_lfifo(0) = '1')then
				lfifo_di(35 downto 34) <= lfifo_di(35 downto 34) or hq(19 downto 18);
			else
				lfifo_di(35 downto 34) <= hq(19 downto 18);
			end if;
		end if;
  end if;
end process;
process(l_clk,cc_rst)
begin
  if(cc_rst = '1')then
    lfifo_wa <= (others => '0');
  elsif(l_clk'event and l_clk = '1')then
    if(strobe_i(0) = '1')then
			if(we_lfifo(3) = '1')then
				lfifo_wa <= lfifo_wa + 1;
			end if;
		end if;
  end if;
end process;
lfifo_di(33) <= '0';
lfifo_di(32) <= '1' when RCV = trlr else '0';
lfifo_di(31 downto 16) <= lfifo_di(15 downto 0);
i_lfifo : RAMB16BWE_S36_S36
   port map (
      DOA => lfifo_do(31 downto 0),      -- Port A 32-bit Data Output
      DOPA => lfifo_do(35 downto 32),    -- Port A 4-bit Parity Output
      ADDRA => lfifo_ra,  -- Port A 9-bit Address Input
      ADDRB => lfifo_wa,  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A 1-bit Clock
      CLKB => l_clk,    -- Port B 1-bit Clock
      DIA => x"00000000",      -- Port A 32-bit Data Input
      DIB => lfifo_di(31 downto 0),      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => lfifo_di(35 downto 32),    -- Port B 4-bit parity Input
--      ENA => strobe,      -- Port A 1-bit RAM Enable Input
      ENA => '1',      -- Port A 1-bit RAM Enable Input
      ENB => strobe_i(0),      -- Port B 1-bit RAM Enable Input
      SSRA => rst_lfifo_do,    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => x"0",      -- Port A 4-bit Write Enable Input
      WEB => we_lfifo       -- Port B 4-bit Write Enable Input
   );
rst_lfifo_do <= not ec_lfifo_ra;
process(clk)
begin
  if(clk'event and clk = '1')then
--		if(strobe = '1')then
			if(cc_rst = '1')then
				wa_sync <= (others => '0');
				up_count <= '0';
				lfifo_ra <= (others => '0');
				lfifo_wc <= (others => '0');
			else
				wa_sync <= wa_sync(1 downto 0) & lfifo_wa(0);
				up_count <= wa_sync(2) xor wa_sync(1);
				if(ec_lfifo_ra = '1')then
					lfifo_ra <= lfifo_ra + 1;
				end if;	
				if(up_count /= ec_lfifo_ra)then
					if(up_count = '1')then
						lfifo_wc <= lfifo_wc + 1;
					else
						lfifo_wc <= lfifo_wc - 1;
					end if;
				end if;
			end if;
--		end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
--		if(strobe = '1')then
			if(cc_rst = '1')then
				lfifo_req_i <= '0';
				CERR <= '0';
				UERR <= '0';
--				BADID <= '0';
--				trailer_i <= '0';
				header <= '1';
				ec_lfifo_ra <= '0';
				lfifo_we_i <= '0';
				blk_length <= "0000001";
				rqst_cntr <= (others => '0');
			else
				if(ec_lfifo_ra = '1')then
					lfifo_req_i <= '0';
				elsif(lfifo_wc(8 downto 6) /= "000" or rqst_cntr(3) = '1' or (header = '1' and lfifo_wc(1 downto 0) = "11"))then
					lfifo_req_i <= '1';
				end if;
				UERR <= lfifo_do(35);
				CERR <= lfifo_do(34);
--				BADID <= lfifo_do(19) and lfifo_do(32);
--				trailer_i <= lfifo_do(32);
				if(lfifo_ack = '1')then
					ec_lfifo_ra <= '1';
				elsif((lfifo_wc(8 downto 1) = x"00" and (lfifo_wc(0) = '0' or ec_lfifo_ra = '1')) or blk_length(6) = '1')then
					ec_lfifo_ra <= '0';
				end if;
				lfifo_we_i <= ec_lfifo_ra;
				if(lfifo_do(32) = '1' and lfifo_we_i = '1')then
					header <= '1';
				elsif(blk_length(5 downto 0) = "000100")then
					header <= '0';
				end if;
--				if(blk_length(6) = '1')then
--					blk_length <= "0000001";
--				elsif(lfifo_do(32) = '1' and lfifo_we_i = '1')then
--					blk_length <= ("000000" & ec_lfifo_ra) + 1;
--				elsif(ec_lfifo_ra = '1')then
--					blk_length <= blk_length + 1;
--				end if;
				if(lfifo_do(32) = '1' and lfifo_we_i = '1')then
					if(ec_lfifo_ra = '1')then
						blk_length <= "0000010";
					else
						blk_length <= "0000001";
					end if;
				elsif(ec_lfifo_ra = '1')then
					if(blk_length(6) = '1')then
						blk_length <= "0000001";
					else
						blk_length <= blk_length + 1;
					end if;
				end if;
				if(or_reduce(lfifo_wc(5 downto 0)) = '0' or up_count = '1' or ec_lfifo_ra = '1')then
					rqst_cntr <= (others => '0');
				else
					rqst_cntr <= rqst_cntr + 1;
				end if;
			end if;
--		end if;
  end if;  
end process;
lfifo_data <= lfifo_do(32 downto 0);
end my_arch;
-------------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
library UNISIM;
use UNISIM.VComponents.all;
entity cntr is
  port (
    reset: in std_logic;
    ec: in std_logic;
    clk: in std_logic;
    q: out std_logic_vector (2 downto 0)
  );
end cntr;

architecture my_arch of cntr is
signal	qi,s,dq:	std_logic_vector(2 downto 0) := (others => '0');
signal	ci:	std_logic_vector(3 downto 0) := (others => '0');
begin
q <= qi;
ci(0) <= not reset;
G1: for I in q'range generate
  s(I) <= qi(I) and not reset;
  mxcy: muxcy    port map (S => s(I), DI => '0', CI => ci(I), O => ci(I+1));
  xocy: xorcy       port map ( LI => s(I), CI => ci(I), O => dq(I));
  ff: fdrse    port map ( C => clk, D => dq(I), CE => ec, S => '0', R => '0', Q => qi(I));
end generate G1;
end my_arch;
