-- MT46V16M16TG-5B
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity sdram_a is
		generic (sim : boolean := false);
    port (
-- reset input
        reset: in STD_LOGIC;
        strobe: in STD_LOGIC;
-- system clock input
        mclk0: in STD_LOGIC; -- 150MHz 
        mclk90: in STD_LOGIC; 
        mclk180: in STD_LOGIC; 
        mclk270: in STD_LOGIC; 
				lock: in std_logic;
-- input interface signals
-- ddr write queue signals
				ddr_rqst: in std_logic;
				ddr_read: in std_logic;
				ddr_gnt: out std_logic;
        ddr_addr: in STD_LOGIC_VECTOR (22 downto 0); -- wr_addr must be all zero when mem_status(14) = '0'
        ddr_length: in STD_LOGIC_VECTOR (5 downto 0);
        wr_data: in STD_LOGIC_VECTOR (31 downto 0);
				wr_en: in std_logic;
-- output interface signals
        rd_en: in STD_LOGIC;
        rd_data: out STD_LOGIC_VECTOR (31 downto 0);
-- ram pads
				mdq : inout STD_LOGIC_VECTOR (15 downto 0);
				maddr : out STD_LOGIC_VECTOR (12 downto 0);
				mba : out STD_LOGIC_VECTOR (1 downto 0);
				dm : out STD_LOGIC_VECTOR (1 downto 0);
				dqs : inout STD_LOGIC_VECTOR (1 downto 0);
				we_n : out STD_LOGIC;
				ras_n : out STD_LOGIC;
				cas_n : out STD_LOGIC;
				cs_n : out STD_LOGIC;
				mclke : out STD_LOGIC;
				mclk : out STD_LOGIC;
				mclk_n : out STD_LOGIC
    );
end sdram_a;

architecture sdram_arch of sdram_a is
COMPONENT mq_cell
	PORT(
		mq_i : IN std_logic;
		dqs : IN std_logic;
		clk : IN std_logic;
		clkn : IN std_logic;
		en_dqs : IN std_logic;
		sel_in : IN std_logic_vector(1 downto 0);          
		sel_out : IN std_logic_vector(1 downto 0);          
		q0 : OUT std_logic;
		q1 : OUT std_logic
		);
END COMPONENT;
-- sdram i/o
signal ma: std_logic_vector(14 downto 0) := (others => '0');
signal md: std_logic_vector(31 downto 0) := (others => '0');
signal mq: std_logic_vector(31 downto 0) := (others => '0');
signal active: std_logic :='0';
signal rasn: std_logic :='0';
signal casn: std_logic :='0';
signal wen: std_logic :='0';
signal active_dl: std_logic_vector(2 downto 0) := (others => '0');
signal dqs_i: std_logic_vector(1 downto 0) := (others => '0');
signal dqs_o: std_logic_vector(1 downto 0) := (others => '0');
signal t_md: std_logic :='0';
signal dqs_outp: std_logic :='0';
signal dqs_out: std_logic :='0';
signal en_dqs: std_logic :='0';
signal d_t_dqs: std_logic :='0';
signal t_dqs: std_logic_vector(1 downto 0) := (others => '0');
signal d_dm: std_logic :='0';
-- refresh signal generation
signal ref: std_logic :='0';
signal start_init: std_logic :='0';
-- sdram state machine
--signal is_read: std_logic :='0';
signal rqst: std_logic :='0';
signal EMR: std_logic := '0';
signal MR: std_logic := '0';
signal w_cycle: std_logic :='0';
signal r_cycle: std_logic :='0';
signal burst_cntr: std_logic_vector(6 downto 0) := (others => '0');
signal wr_burst: std_logic :='0';
signal wr_burst_dl: std_logic :='0';
signal rd_burst: std_logic :='0';
signal rd_burst_dl: std_logic :='0';
signal rd_burst_dl7: std_logic :='0';
signal tRAS: std_logic :='0';
signal stop_r: std_logic :='0';
signal stop_w: std_logic :='0';
signal inh_ras: std_logic :='0';
signal ref_rqst: std_logic :='0';
signal init_done: std_logic_vector(1 downto 0) := (others => '0');
signal pc_dl: std_logic :='0';
signal precharge: std_logic :='0';
signal precharge_dl: std_logic_vector(1 downto 0) := (others => '0');
signal rfc_cntr: std_logic_vector(4 downto 0) := (others => '0');
signal wfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_dout : std_logic_vector(31 downto 0) := (others => '0');
signal rfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_we: std_logic := '0';
signal en_out: std_logic := '0';
signal sel_in: std_logic_vector(3 downto 0) := (others => '0');
signal sel_out: std_logic_vector(1 downto 0) := (others => '0');
signal mdq_in : std_logic_vector(15 downto 0) := (others => '0');
signal mdq_out : std_logic_vector(15 downto 0) := (others => '0');
signal ddr_gntp: std_logic := '0';
attribute clock_signal : string;
attribute clock_signal of dqs_i : signal is "yes";
begin
i_mclk:  oddr2
    port map(
          d0 => '1',
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk);
i_mclk_n:  oddr2
    port map(
          d0 => '0',
          d1 => '1',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk_n);
-- ram pads
i_dqs_o0:  oddr2
    port map(
          d0 => dqs_out,
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => dqs_o(0));
i_dqs0 : IOBUF
   port map (
      O => dqs_i(0),     -- Buffer output
      IO => dqs(0),   -- Buffer inout port (connect directly to top-level port)
      I => dqs_o(0),     -- Buffer input
      T => t_dqs(0)      -- 3-state enable input 
   );
i_dqs_o1:  oddr2
    port map(
          d0 => dqs_out,
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => dqs_o(1));
i_dqs1 : IOBUF
   port map (
      O => dqs_i(1),     -- Buffer output
      IO => dqs(1),   -- Buffer inout port (connect directly to top-level port)
      I => dqs_o(1),     -- Buffer input
      T => t_dqs(1)      -- 3-state enable input 
   );
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    we_n <= wen;
    cas_n <= casn;
    ras_n <= rasn;
    maddr <= ma(12 downto 0);
		mba <= ma(14 downto 13);
  end if;
end process;
md(15 downto 0) <= wfifo_dout(15 downto 0);
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
    md(31 downto 16) <= wfifo_dout(31 downto 16);
  end if;
end process;
i_wfifo : RAMB16_S36_S36
   port map (
      DOB => wfifo_dout,      -- Port B 32-bit Data Output
      ADDRA => wfifo_wa,  -- Port A 9-bit Address Input
      ADDRB => wfifo_ra,  -- Port B 9-bit Address Input
      CLKA => mclk90,    -- Port A 1-bit Clock
      CLKB => mclk180,    -- Port B 1-bit Clock
      DIA => wr_data,      -- Port A 32-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
--      ENA => strobe,      -- Port A 1-bit RAM Enable Input
      ENA => '1',      -- Port A 1-bit RAM Enable Input
      ENB => '1',      -- Port B 1-bit RAM Enable Input
      SSRA => '0',    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => wr_en,      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );
process(mclk180,reset)
begin
  if(reset = '1')then
	  wfifo_ra <= (others => '0');
  elsif(mclk180'event and mclk180 = '1')then
	  if(wr_burst_dl = '1')then
		  wfifo_ra <= wfifo_ra + 1;
		end if;
  end if;
end process;  
i_rfifo : RAMB16_S36_S36
   port map (
      DOB => rd_data,      -- Port B 32-bit Data Output
      ADDRA => rfifo_wa,  -- Port A 9-bit Address Input
      ADDRB => rfifo_ra,  -- Port B 9-bit Address Input
      CLKA => mclk0,    -- Port A 1-bit Clock
      CLKB => mclk90,    -- Port B 1-bit Clock
      DIA => mq,      -- Port A 32-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => rfifo_we,      -- Port A 1-bit RAM Enable Input
      ENB => strobe,      -- Port B 1-bit RAM Enable Input
      SSRA => '0',    -- Port A 1-bit Synchronous Set/Reset Input
      SSRB => '0',    -- Port B 1-bit Synchronous Set/Reset Input
      WEA => '1',      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );
process(mclk90)
begin
  if(mclk90'event and mclk90 = '1')then
--		if(strobe = '1')then
			if(reset = '1')then
				wfifo_wa <= (others => '0');
			elsif(wr_en = '1')then
				wfifo_wa <= wfifo_wa + 1;
			end if;
		if(strobe = '1')then
			if(reset = '1')then
				rfifo_ra <= (others => '0');
			elsif(rd_en = '1')then
				rfifo_ra <= rfifo_ra + 1;
			end if;
			ddr_gnt <= ddr_gntp;
		end if;
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(reset = '1')then
      rfifo_we <= '0';
		else
			rfifo_we <= rd_burst_dl7;
		end if;
    if(reset = '1')then
      rfifo_wa <= (others => '0');
		elsif(rfifo_we = '1')then
			rfifo_wa <= rfifo_wa + 1;
		end if;
  end if;
end process;
i_rd_burst_dl: delayN
  generic map(N=> 3) port map(d => rd_burst, ce => '1', c => mclk0, q => rd_burst_dl);
i_rd_burst_dl7: delayN
  generic map(N=> 7) port map(d => rd_burst, ce => '1', c => mclk0, q => rd_burst_dl7);
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
		dm <= d_dm & d_dm;
		ddr_gntp <= active_dl(0) or active_dl(1);
  end if; 
end process;
-- ram access request logic
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(reset = '1' or precharge = '1' or stop_r = '1')then
      rqst <= '0';
    elsif(ddr_rqst = '1')then
      rqst <= '1';
    end if;
--    if(rqst = '0')then
--			is_read <= ddr_read;
--    end if;
-- this ensures active, read/write and precharge are to the same bank
		if(init_done(1) = '0')then
			ma(14 downto 13) <= '0' & EMR;
    elsif(inh_ras = '0')then
			ma(14 downto 13) <= ddr_addr(22 downto 21);
		end if;
		if(init_done(1) = '0')then
-- latency = 3 burst = 8
			ma(12 downto 0) <= "00" & precharge & '0' & MR & "00" & MR & MR & "00" & (MR or EMR) & MR; 
		else
			if(inh_ras = '0')then
				ma(12 downto 0) <= ddr_addr(20 downto 8);
			else
				if(active_dl(1) = '1')then
					ma(8 downto 0) <= ddr_addr(7 downto 0) & '0';
				elsif(burst_cntr(6) = '1')then 	
					ma(6 downto 1) <= ma(6 downto 1) + 1;
				end if;
--				ma(12 downto 9) <= x"0";				
				ma(12 downto 9) <= "00" & precharge & '0';				
			end if;
		end if;
		rd_burst <= burst_cntr(6) and r_cycle;
		wr_burst <= burst_cntr(6) and w_cycle;
		wr_burst_dl <= wr_burst;
		active_dl <= active_dl(1 downto 0) & active;
  end if;
end process;
-- generate read and write burst signals -------------------------------------------------------
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(reset = '1')then
			t_md <= '1';
			d_dm <= '0';
			burst_cntr <= (others => '0');
		else
			t_md <= not(not casn and w_cycle) and (t_md or pc_dl);
			d_dm <= stop_w or (d_dm and casn);
			if(active_dl(0) = '1')then
				burst_cntr <= '1' & not ddr_length;
			elsif(burst_cntr(6) = '1')then
				burst_cntr <= burst_cntr + 1;
			end if;
		end if;
  end if;
end process;
process(mclk270,reset)
begin
	if(reset = '1')then
		en_dqs <= '0';
		d_t_dqs <= '1';
		t_dqs <= "11";
		dqs_outp <= '0';
		dqs_out <= '0';
  elsif(mclk270'event and mclk270 = '1')then
		en_dqs <= not active_dl(2) and (en_dqs or rd_burst_dl);
		d_t_dqs <= not active_dl(0) and (d_t_dqs or precharge_dl(0) or rd_burst_dl);
		t_dqs <= d_t_dqs & d_t_dqs;
		dqs_outp <= (not casn and w_cycle) or (dqs_outp and not precharge);
		dqs_out <= dqs_outp;
	end if;
end process;
-- sdram state machine -----------------------------------------------------------------------------------------
-- sdram power on initialization states: precharge all banks, enable DLL, reset DLL, precharge all banks,
-- two refresh cycle and set mode register.
i_EMR: delayN
  generic map(N=> 5) port map(d => start_init, ce => '1', c => mclk0, q => EMR);
i_MR: delayN
  generic map(N=> 7) port map(d => start_init, ce => '1', c => mclk0, q => MR);
process(mclk0,lock)
variable state: std_logic_vector(2 downto 0);
begin
state := rasn & casn & wen;
  if(lock = '0')then
    w_cycle <= '0';
    r_cycle <= '0';
    pc_dl <= '0';
    precharge <= '0';
    precharge_dl <= (others => '0');
    rfc_cntr <= (others => '0');
    tRAS <= '0';
		inh_ras <= '0';
    active <= '0';
    stop_w <= '0';
    stop_r <= '0';
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    init_done <= "00";
    ref_rqst <= '0';
  elsif(mclk0'event and mclk0 = '1')then
    if(active = '1')then
			tRAS <= '1';
    elsif(active_dl(1) = '1')then
			tRAS <= '0';
		end if;
		stop_w <= wr_burst_dl and not wr_burst;
		pc_dl <= start_init or MR or stop_w or stop_r;
    precharge <= pc_dl;
    precharge_dl <= precharge_dl(0) & precharge;
-- auto refresh tRFC counter
    if(state = "001")then
			rfc_cntr <= "10110";
		elsif(rfc_cntr(4) = '1')then
			rfc_cntr <= rfc_cntr + 1;
		end if;
-- inhibit RAS after refresh or read/write until tRFC or tPR
    if(rasn = '0' and wen = '1')then
			inh_ras <= '1';
    elsif(precharge_dl(1) = '1' or rfc_cntr(3 downto 1) = "111")then
			inh_ras <= '0';
		end if;
-- latch refresh request
    if(ref = '1')then
      ref_rqst <= '1';
    end if;
		active <= '0';
		if(r_cycle = '1' and stop_r = '0' and (burst_cntr(6) = '0' or and_reduce(burst_cntr(5 downto 1)) = '1') and tRAS = '0')then
			stop_r <= '1';
		else
			stop_r <= '0';
		end if;
    if(active_dl(0) = '1')then
--			r_cycle <= is_read;
			r_cycle <= ddr_read;
		elsif(stop_r = '1')then
			r_cycle <= '0';
		end if;
    if(active_dl(0) = '1')then
--			w_cycle <= not is_read;
			w_cycle <= not ddr_read;
		elsif(stop_w = '1')then
			w_cycle <= '0';
		end if;
    case state is
-- NOP
      when "111" =>
---- goto precharge
        if(precharge = '1')then
          rasn <= '0';
          wen <= '0';
        end if;
---- goto auto refresh
        if(ref_rqst = '1' and inh_ras = '0')then
          rasn <= '0';
          casn <= '0';
        end if;
---- goto set mode
        if(EMR = '1' or MR = '1')then
          rasn <= '0';
          casn <= '0';
          wen <= '0';
        end if;
---- goto active
        if(rqst = '1' and ref_rqst = '0' and inh_ras = '0')then
          rasn <= '0';
					active <= '1';
        end if;
------ goto write/read
        if(active_dl(1) = '1' or (burst_cntr(6) = '1' and ma(2 downto 1) = "11"))then
          casn <= '0';
          wen <= r_cycle;
        end if;
-- AUTO REFRESH
      when "001" =>
				ref_rqst <= '0';
        init_done <= init_done(0) & '1';
				rasn <= '1';
				casn <= '1';
      when others =>
        if(burst_cntr(6) = '0' or ma(2 downto 1) /= "11")then
-- default to go to NOP
					rasn <= '1';
					casn <= '1';
					wen <= '1';
        end if;
    end case;
  end if;
end process;
----------------------------------------------------------------------------------------
g_mdq: for i in 0 to 15 generate
	i_mdq_out:  oddr2
    port map(
          d0 => md(16+i),
          d1 => md(i),
          c0 => mclk0,
          c1 => mclk180,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mdq_out(i));
	i_mdq : IOBUF
   port map (
      O => mdq_in(i),     -- Buffer output
      IO => mdq(i),   -- Buffer inout port (connect directly to top-level port)
      I => mdq_out(i),     -- Buffer input
      T => t_md      -- 3-state enable input 
   );
end generate g_mdq;
process(dqs_i,en_dqs)
begin
	if(en_dqs = '0')then
		sel_in(1) <= '0';
	elsif(dqs_i(0)'event and dqs_i(0) = '1')then
		sel_in(1) <= not sel_in(1);
	end if;
end process;
process(dqs_i,en_dqs)
begin
	if(en_dqs = '0')then
		sel_in(0) <= '0';
	elsif(dqs_i(0)'event and dqs_i(0) = '0')then
		sel_in(0) <= not sel_in(0);
	end if;
end process;
process(dqs_i,en_dqs)
begin
	if(en_dqs = '0')then
		sel_in(3) <= '0';
	elsif(dqs_i(1)'event and dqs_i(1) = '1')then
		sel_in(3) <= not sel_in(3);
	end if;
end process;
process(dqs_i,en_dqs)
begin
	if(en_dqs = '0')then
		sel_in(2) <= '0';
	elsif(dqs_i(1)'event and dqs_i(1) = '0')then
		sel_in(2) <= not sel_in(2);
	end if;
end process;
g_mq_cell: for i in 0 to 15 generate
  i_mq_cell: mq_cell PORT MAP(
		mq_i => mdq_in(i),
		dqs => dqs_i(i/8),
		clk => mclk0,
		clkn => mclk180,
		en_dqs => en_dqs,
		sel_in => sel_in(2*(i/8)+1 downto 2*(i/8)),
		sel_out => sel_out,
		q0 => mq(i),
		q1 => mq(i+16)
	);
end generate;
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
		sel_out(0) <= sel_out(1);
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(en_out = '0')then
			sel_out(1) <= '0';
		else
			sel_out(1) <= not sel_out(1);
		end if;
  end if;
end process;
i_en_out: delayN
  generic map(N=> 5) port map(d => rd_burst, ce => '1', c => mclk0, q => en_out);
----------------------------------------------------------------------------------------
-- instantiate refresh generator
I_refresh: refresh
	generic map(sim => sim, N => 1)
	PORT MAP(
		clk => mclk0,
		lock => lock,
		init_done => init_done(1),
		start_init => start_init,
		ddr_csn => cs_n,
		ddr_clke => mclke,
		ref => ref
	);
end sdram_arch;
